"""
=======================
带有混合类型的列转换器
=======================

.. currentmodule:: sklearn

本示例说明如何使用 :class:`~compose.ColumnTransformer` 对不同特征子集应用不同的预处理和特征提取管道。这对于包含异构数据类型的数据集特别有用，因为我们可能希望对数值特征进行缩放，对分类特征进行独热编码。

在此示例中，数值数据在均值插补后进行标准缩放。分类数据通过 ``OneHotEncoder`` 进行独热编码，该编码器为缺失值创建了一个新类别。我们进一步通过卡方检验选择类别来减少维度。

此外，我们展示了两种将列分派给特定预处理器的方法：按列名和按列数据类型。

最后，预处理管道与一个简单的分类模型一起集成在一个完整的预测管道中，使用 :class:`~pipeline.Pipeline` 。

"""

# 作者：scikit-learn 开发者
# SPDX 许可证标识符：BSD-3-Clause

# %%
import numpy as np

from sklearn.compose import ColumnTransformer
from sklearn.datasets import fetch_openml
from sklearn.feature_selection import SelectPercentile, chi2
from sklearn.impute import SimpleImputer
from sklearn.linear_model import LogisticRegression
from sklearn.model_selection import RandomizedSearchCV, train_test_split
from sklearn.pipeline import Pipeline
from sklearn.preprocessing import OneHotEncoder, StandardScaler

np.random.seed(0)

# %%
# 从 https://www.openml.org/d/40945 加载数据

X, y = fetch_openml("titanic", version=1, as_frame=True, return_X_y=True)

# 或者，可以直接从frame属性中获取X和y：
# X = titanic.frame.drop('survived', axis=1)
# y = titanic.frame['survived']

# %%
# 使用 ``ColumnTransformer`` 通过列名选择列
#
# 我们将使用以下特征来训练我们的分类器：
#
# Numeric Features:
#
# * ``age`` : 浮点数；
# * ``fare`` : 浮点数。
#
# 类别特征:
#
# * ``embarked`` ：类别编码为字符串 ``{'C', 'S', 'Q'}`` ;
# * ``sex`` ：类别编码为字符串 ``{'female', 'male'}`` ;
# * ``pclass`` ：有序整数 ``{1, 2, 3}`` 。
#
# 我们为数值数据和分类数据创建预处理管道。注意， ``pclass`` 可以被视为分类特征或数值特征。

numeric_features = ["age", "fare"]
numeric_transformer = Pipeline(
    steps=[("imputer", SimpleImputer(strategy="median")), ("scaler", StandardScaler())]
)

categorical_features = ["embarked", "sex", "pclass"]
categorical_transformer = Pipeline(
    steps=[
        ("encoder", OneHotEncoder(handle_unknown="ignore")),
        ("selector", SelectPercentile(chi2, percentile=50)),
    ]
)
preprocessor = ColumnTransformer(
    transformers=[
        ("num", numeric_transformer, numeric_features),
        ("cat", categorical_transformer, categorical_features),
    ]
)

# %%
# 将分类器附加到预处理管道。
# 现在我们有一个完整的预测管道。

clf = Pipeline(
    steps=[("preprocessor", preprocessor), ("classifier", LogisticRegression())]
)

X_train, X_test, y_train, y_test = train_test_split(X, y, test_size=0.2, random_state=0)

clf.fit(X_train, y_train)
print("model score: %.3f" % clf.score(X_test, y_test))

# %%
# ``Pipeline`` 的 HTML 表示（显示图表）
#
# 当在 jupyter notebook 中打印 ``Pipeline`` 时，会显示估计器的 HTML 表示形式：
clf

# %%
# 使用 ``ColumnTransformer`` 按数据类型选择列
#
# 在处理清理过的数据集时，可以通过使用列的数据类型来自动进行预处理，以决定是否将某列视为数值特征或类别特征。:func:`sklearn.compose.make_column_selector` 提供了这种可能性。首先，让我们只选择一部分列来简化我们的示例。

subset_feature = ["embarked", "sex", "pclass", "age", "fare"]
X_train, X_test = X_train[subset_feature], X_test[subset_feature]

# %%
# 然后，我们内省每列数据类型的信息。


X_train.info()

# %%
# 我们可以观察到，在使用 ``fetch_openml`` 加载数据时， `embarked` 和 `sex` 列被标记为 `category` 列。因此，我们可以使用这些信息将分类列分派给 ``categorical_transformer`` ，而将其余列分派给 ``numerical_transformer`` 。


# %%
# .. NOTE:: 实际操作中，你需要自己处理列的数据类型。如果你希望某些列被视为 `category` ，你需要将它们转换为分类列。如果你使用的是pandas，可以参考他们关于 `分类数据 <https://pandas.pydata.org/pandas-docs/stable/user_guide/categorical.html>`_  的文档。


from sklearn.compose import make_column_selector as selector

preprocessor = ColumnTransformer(
    transformers=[
        ("num", numeric_transformer, selector(dtype_exclude="category")),
        ("cat", categorical_transformer, selector(dtype_include="category")),
    ]
)
clf = Pipeline(
    steps=[("preprocessor", preprocessor), ("classifier", LogisticRegression())]
)


clf.fit(X_train, y_train)
print("model score: %.3f" % clf.score(X_test, y_test))
clf

# %%
# 最终得分与之前的流水线不完全相同，因为基于数据类型的选择器将“pclass”列视为数值特征，而不是之前的分类特征：

selector(dtype_exclude="category")(X_train)

# %%

selector(dtype_include="category")(X_train)

# %%
# 在网格搜索中使用预测管道
#
# 网格搜索也可以在 ``ColumnTransformer`` 对象中定义的不同预处理步骤上执行，并将其与分类器的超参数一起作为 ``Pipeline`` 的一部分进行搜索。我们将使用:class:`~sklearn.model_selection.RandomizedSearchCV` 搜索数值预处理的插补策略和逻辑回归的正则化参数。此超参数搜索会随机选择由 `n_iter` 配置的固定数量的参数设置。或者，也可以使用:class:`~sklearn.model_selection.GridSearchCV` ，但这将评估参数空间的笛卡尔积。

param_grid = {
    "preprocessor__num__imputer__strategy": ["mean", "median"],
    "preprocessor__cat__selector__percentile": [10, 30, 50, 70],
    "classifier__C": [0.1, 1.0, 10, 100],
}

search_cv = RandomizedSearchCV(clf, param_grid, n_iter=10, random_state=0)
search_cv

# %%
# 调用 'fit' 会触发交叉验证搜索最佳超参数组合：
#
search_cv.fit(X_train, y_train)

print("Best params:")
print(search_cv.best_params_)

# %%
# 内部交叉验证得分如下：

print(f"Internal CV score: {search_cv.best_score_:.3f}")

# %%
# 我们还可以将顶级网格搜索结果作为 pandas 数据框进行自省：

import pandas as pd

cv_results = pd.DataFrame(search_cv.cv_results_)
cv_results = cv_results.sort_values("mean_test_score", ascending=False)
cv_results[
    [
        "mean_test_score",
        "std_test_score",
        "param_preprocessor__num__imputer__strategy",
        "param_preprocessor__cat__selector__percentile",
        "param_classifier__C",
    ]
].head(5)

# %%
# 最佳超参数已用于在完整训练集上重新拟合最终模型。我们可以在未用于超参数调优的保留测试数据上评估该最终模型。
#
print(
    "accuracy of the best model from randomized search: "
    f"{search_cv.score(X_test, y_test):.3f}"
)
