r"""
=====================================================================
使用随机投影进行嵌入的Johnson-Lindenstrauss界限
=====================================================================

 `Johnson-Lindenstrauss引理`_  指出，任何高维数据集都可以随机投影到低维欧几里得空间，同时控制成对距离的失真。

.. _ `Johnson-Lindenstrauss引理` : https://en.wikipedia.org/wiki/\
    Johnson%E2%80%93Lindenstrauss_lemma

"""

import sys
from time import time

import matplotlib.pyplot as plt
import numpy as np

from sklearn.datasets import fetch_20newsgroups_vectorized, load_digits
from sklearn.metrics.pairwise import euclidean_distances
from sklearn.random_projection import (
    SparseRandomProjection,
    johnson_lindenstrauss_min_dim,
)

# %%
# 理论界限
# ========
# 随机投影 `p` 引入的失真由以下事实断言： `p` 以较高概率定义了一个 eps-嵌入，如下所示：
#
# .. math::
# (1 - eps) \|u - v\|^2 < \|p(u) - p(v)\|^2 < (1 + eps) \|u - v\|^2
#
# 其中 `u` 和 `v` 是从形状为 `(n_samples, n_features)` 的数据集中提取的任意行， `p` 是通过形状为 `(n_components, n_features)` 的随机高斯 `N(0, 1)` 矩阵（或稀疏 Achlioptas 矩阵）进行的投影。
#
# 保证eps-嵌入的最小组件数由以下公式给出：
#
# .. math::
# n\_components \geq 4 \log(n\_samples) / \left(\frac{eps^2}{2} - \frac{eps^3}{3}\right)
#
#
# 第一个图表显示，随着样本数量 ``n_samples`` 的增加，为了保证 ``eps`` -嵌入，最小维数 ``n_components`` 以对数方式增加。

# 可接受的失真范围
eps_range = np.linspace(0.1, 0.99, 5)
colors = plt.cm.Blues(np.linspace(0.3, 1.0, len(eps_range)))

# 样本（观测）数量的范围以嵌入
n_samples_range = np.logspace(1, 9, 9)

plt.figure()
for eps, color in zip(eps_range, colors):
    min_n_components = johnson_lindenstrauss_min_dim(n_samples_range, eps=eps)
    plt.loglog(n_samples_range, min_n_components, color=color)

plt.legend([f"eps = {eps:0.1f}" for eps in eps_range], loc="lower right")
plt.xlabel("Number of observations to eps-embed")
plt.ylabel("Minimum number of dimensions")
plt.title("Johnson-Lindenstrauss bounds:\nn_samples vs n_components")
plt.show()


# %%
# 第二个图表显示，增加允许的失真度“eps”可以显著减少给定样本数量“n_samples”所需的最小维数“n_components”。


# 可接受的失真范围
eps_range = np.linspace(0.01, 0.99, 100)

# 样本（观测）数量的范围以嵌入
n_samples_range = np.logspace(2, 6, 5)
colors = plt.cm.Blues(np.linspace(0.3, 1.0, len(n_samples_range)))

plt.figure()
for n_samples, color in zip(n_samples_range, colors):
    min_n_components = johnson_lindenstrauss_min_dim(n_samples, eps=eps_range)
    plt.semilogy(eps_range, min_n_components, color=color)

plt.legend([f"n_samples = {n}" for n in n_samples_range], loc="upper right")
plt.xlabel("Distortion eps")
plt.ylabel("Minimum number of dimensions")
plt.title("Johnson-Lindenstrauss bounds:\nn_components vs eps")
plt.show()

# %%
# 经验验证
# ========
#
# 我们在20个新闻组文本文档（TF-IDF词频）数据集或数字数据集上验证上述界限：
#
# - 对于20个新闻组数据集，使用稀疏随机矩阵将总共包含10万个特征的大约300个文档投影到较小的欧几里得空间中，目标维度数量的不同值由 ``n_components`` 指定。
#
# 对于数字数据集，将300张手写数字图片的8x8灰度像素数据随机投影到不同数量的更高维空间 ``n_components`` 。
#
# 默认数据集是20个新闻组数据集。要在数字数据集上运行此示例，请向此脚本传递 ``--use-digits-dataset`` 命令行参数。

if "--use-digits-dataset" in sys.argv:
    data = load_digits().data[:300]
else:
    data = fetch_20newsgroups_vectorized().data[:300]

# %%
# 对于每个 ``n_components`` 的值，我们绘制：
#
# - 样本对的二维分布，其中原始空间和投影空间中的成对距离分别作为 x 轴和 y 轴。
#
# - 投影距离与原始距离比值的一维直方图。

n_samples, n_features = data.shape
print(
    f"Embedding {n_samples} samples with dim {n_features} using various "
    "random projections"
)

n_components_range = np.array([300, 1_000, 10_000])
dists = euclidean_distances(data, squared=True).ravel()

# 选择仅非相同的样本对
nonzero = dists != 0
dists = dists[nonzero]

for n_components in n_components_range:
    t0 = time()
    rp = SparseRandomProjection(n_components=n_components)
    projected_data = rp.fit_transform(data)
    print(
        f"Projected {n_samples} samples from {n_features} to {n_components} in "
        f"{time() - t0:0.3f}s"
    )
    if hasattr(rp, "components_"):
        n_bytes = rp.components_.data.nbytes
        n_bytes += rp.components_.indices.nbytes
        print(f"Random matrix with size: {n_bytes / 1e6:0.3f} MB")

    projected_dists = euclidean_distances(projected_data, squared=True).ravel()[nonzero]

    plt.figure()
    min_dist = min(projected_dists.min(), dists.min())
    max_dist = max(projected_dists.max(), dists.max())
    plt.hexbin(
        dists,
        projected_dists,
        gridsize=100,
        cmap=plt.cm.PuBu,
        extent=[min_dist, max_dist, min_dist, max_dist],
    )
    plt.xlabel("Pairwise squared distances in original space")
    plt.ylabel("Pairwise squared distances in projected space")
    plt.title("Pairwise distances distribution for n_components=%d" % n_components)
    cb = plt.colorbar()
    cb.set_label("Sample pairs counts")

    rates = projected_dists / dists
    print(f"Mean distances rate: {np.mean(rates):.2f} ({np.std(rates):.2f})")

    plt.figure()
    plt.hist(rates, bins=50, range=(0.0, 2.0), edgecolor="k", density=True)
    plt.xlabel("Squared distances rate: projected / original")
    plt.ylabel("Distribution of samples pairs")
    plt.title("Histogram of pairwise distance rates for n_components=%d" % n_components)

    # 待办事项：计算 eps 的期望值，并将其作为垂直线/区域添加到之前的图中

plt.show()


# %%
# 我们可以看到，对于较低的 ``n_components`` 值，分布较宽，有许多失真的对，并且分布偏斜（由于左侧的比率硬限制为零，因为距离总是正值），而对于较高的 ``n_components`` 值，失真得到控制，随机投影很好地保留了距离。
#
# Remarks
# =======
#
# 根据JL引理，将300个样本投影而不产生过多失真将需要至少几千个维度，而不管原始数据集的特征数量。
#
# 因此，在只有64个输入特征的数字数据集上使用随机投影没有意义：在这种情况下，它不允许进行降维。
#
# 另一方面，在20个新闻组数据集上，维度可以从56,436减少到10,000，同时合理地保留成对距离。
