"""
=====================================
MNIST上MLP权重的可视化
=====================================

有时查看神经网络的学习系数可以提供对学习行为的洞察。例如，如果权重看起来没有结构化，可能有些权重根本没有被使用，或者如果存在非常大的系数，可能正则化太低或学习率太高。

这个例子展示了如何绘制在MNIST数据集上训练的MLPClassifier的第一层权重。

输入数据由28x28像素的手写数字组成，导致数据集中有784个特征。因此，第一层权重矩阵的形状为(784, hidden_layer_sizes[0])。因此，我们可以将权重矩阵的单列可视化为一个28x28像素的图像。

为了使示例运行得更快，我们使用了非常少的隐藏单元，并且只训练了很短的时间。更长时间的训练会导致权重具有更平滑的空间外观。由于没有收敛，示例会抛出一个警告，在这种情况下，这是我们希望的，因为我们的持续集成基础设施上有资源使用限制，该基础设施用于定期构建此文档。
"""

import warnings

import matplotlib.pyplot as plt

from sklearn.datasets import fetch_openml
from sklearn.exceptions import ConvergenceWarning
from sklearn.model_selection import train_test_split
from sklearn.neural_network import MLPClassifier

# 从 https://www.openml.org/d/554 加载数据
# 
# 
X, y = fetch_openml("mnist_784", version=1, return_X_y=True, as_frame=False)
X = X / 255.0

# 将数据分为训练集和测试集
X_train, X_test, y_train, y_test = train_test_split(X, y, random_state=0, test_size=0.7)

mlp = MLPClassifier(
    hidden_layer_sizes=(40,),
    max_iter=8,
    alpha=1e-4,
    solver="sgd",
    verbose=10,
    random_state=1,
    learning_rate_init=0.2,
)

# 这个示例不会收敛，因为我们的持续集成基础设施上的资源使用受到限制，所以我们在这里捕获警告并忽略它。
with warnings.catch_warnings():
    warnings.filterwarnings("ignore", category=ConvergenceWarning, module="sklearn")
    mlp.fit(X_train, y_train)

print("Training set score: %f" % mlp.score(X_train, y_train))
print("Test set score: %f" % mlp.score(X_test, y_test))

fig, axes = plt.subplots(4, 4)
# 使用全局最小值/最大值来确保所有权重在同一尺度上显示
vmin, vmax = mlp.coefs_[0].min(), mlp.coefs_[0].max()
for coef, ax in zip(mlp.coefs_[0].T, axes.ravel()):
    ax.matshow(coef.reshape(28, 28), cmap=plt.cm.gray, vmin=0.5 * vmin, vmax=0.5 * vmax)
    ax.set_xticks(())
    ax.set_yticks(())

plt.show()
